from .vectorizers import DocVectorizer
from sklearn.svm import SVC


class BaseClassifier(object):

    def __init__(self, vectorizer=DocVectorizer()):

        self.vectorizer = vectorizer

    def fit(self, documents):
        """
        :param documents: a dict containing text as keys and label as values
        """

        self.text_vectors = self.vectorizer.fit_transform(list(documents.keys()))

    def predict(self, documents):
        pass


class SVMClassifier(BaseClassifier):

    def fit(self, documents):
        BaseClassifier.fit(self, documents)

        self.sk_model = SVC()



from .tokenizers import CharTokenizer
from .vectorizers import WordVectorizer


class Text(object):
    """
    A TextBlob `Blob`-like class.

    Examples
    --------
    >>> Text('This is a-text !', tokenizer=CharTokenizer()).tokens
    ['This', 'is', 'a', 'text']
    >>> Text('This is another text !') + " " + "So fun"
    'This is another text ! So fun'
    """

    def __init__(self, text, tokenizer=CharTokenizer(), vectorizer=WordVectorizer()):

        self.raw_text = text
        self.words = self.tokens = tokenizer.tokenize(text)
        self.vectorizer = vectorizer

    def __str__(self):
        """
        Examples
        --------
        >>> str(Text('This is a-text !'))
        'This is a-text !'
        """
        return self.raw_text

    def __repr__(self):
        """
        Examples
        --------
        >>> Text('This is a-text !')
        Text('This is a-text !')
        """
        return 'Text({})'.format(repr(self.raw_text))

    def __add__(self, other):
        return self.raw_text + other

    def vectorize(self, window=50, reduce_rate=None):
        self.vectorizer.window = window
        vectors = self.vectorizer.fit_transform(self.raw_text)

        if isinstance(reduce_rate, int):
            return vectors.reduce(reduce_rate)
        else:
            return vectors

    def ngrams(self, n=2):
        """
        Return n-grams of the text. Hazardously found in the locallyoptimal.com blog.
        :param n: length of grams

        Examples
        --------
        >>> Text('This is a-text !').ngrams(n=2)
        [('This', 'is'), ('is', 'a'), ('a', 'text')]
        """
        return list(zip(*[self.tokens[i:] for i in range(n)]))



from .tokenizers import CharTokenizer
from .core import VectorDictionary
import numpy as np


class BaseVectorizer(object):
    """
    A base class for building vectorizers with a scikit-learn like API.
    """

    def __init__(self, window=5, tokenizer=CharTokenizer()):
        """
        :param tokenizer: A BaseTokenizer subclass object to tokenize the text
        """

        self.tokenizer = tokenizer
        self.window = window

        self.raw_documents = []  # List of raw documents
        self.processed_documents = []  # List of tokenized, lowercased documents
        self.unique_words = []  # List of uniques words in `self.document_corpus`

    def fit(self, documents):
        if isinstance(documents, list):
            self.raw_documents += documents
            for doc in documents:
                self.processed_documents.append(self.tokenizer.tokenize(doc.lower()))

        else:
            self.raw_documents.append(documents)
            self.processed_documents.append(self.tokenizer.tokenize(documents.lower()))

        documents_corpus = self.tokenizer.tokenize(" ".join(self.raw_documents).lower())
        self.unique_words = list(set(documents_corpus))

    def transform(self):
        pass

    def fit_transform(self, documents):
        self.fit(documents)
        return self.transform()

    def predict(self, documents):
        pass


class WordVectorizer(BaseVectorizer):
    """
    A simple distributional vectorizer algorithm.
    """

    def transform(self):
        vector_dict = VectorDictionary(dimension=len(self.unique_words))

        for n_word in self.unique_words:
            n_word_vectors = np.empty([0, len(self.unique_words)], dtype=int)

            for document in self.processed_documents:
                doc_n_word_indexes = [i for i, x in enumerate(document) if x == n_word]  # list of index of `n_word` in document

                if doc_n_word_indexes:
                    # Build a vector for each index...
                    for index in doc_n_word_indexes:
                        text_selection = document[index - self.window:index] +\
                                         document[index + 1:index + self.window + 1]

                        n_word_vectors = np.append(n_word_vectors,
                                                   np.array([[text_selection.count(word) for word in self.unique_words]]),
                                                   axis=0)

                else:
                    np.append(n_word_vectors, np.zeros(len(self.unique_words)))

            # ...And sum them to build the final vector :
            vector_dict[n_word] = np.sum(n_word_vectors, axis=0)

        return vector_dict


class DocVectorizer(BaseVectorizer):
    """
    Use WordVectorizer to vectorize a whole text.
    """

    def transform(self):
        vector_dict = VectorDictionary(dimension=len(self.unique_words))

        for document in self.raw_documents:
            vectorizer = WordVectorizer(window=self.window, tokenizer=self.tokenizer)
            vectorizer.fit(document)
            vectorizer.unique_words = self.unique_words

            document_vocabulary_vectors = vectorizer.transform()
            vector_dict[document] = np.mean(np.array(document_vocabulary_vectors.values()), axis=0)

        return vector_dict



from collections import OrderedDict, Counter
import numpy as np
from scipy import spatial
from sklearn.decomposition import TruncatedSVD


class VectorDictionary(object):
    """
    A dictionary that contain coordinates of a named document in vectorization (of words or text).
    """

    def __init__(self, vectors=None, dimension=3):
        """
        :param vectors: pre-filled dictionary
        :param dimension: length of all dictionary values
        """

        self.dimension = dimension

        if vectors is not None:
            for key, coordinates in vectors.items():
                if not isinstance(coordinates, np.ndarray):  # Check if pre-filled dictionary values are numpy arrays.
                    raise TypeError('`vectors` argument should contain numpy arrays values.')

                if not coordinates.size == dimension:  # Check if pre-filled dictionary respect `dimension` argument.
                    raise ValueError('`vectors` argument must contain values with a length that'
                                     ' should be equal to {0} not {1}'.format(self.dimension, coordinates.size))

            self.vectors = OrderedDict(vectors)
        else:
            self.vectors = OrderedDict()

    def __str__(self):
        return "VectorDictionary({})".format(self.items())

    def __repr__(self):
        return "VectorDictionary({})".format(repr(self.vectors))

    def __getitem__(self, key):
        return self.vectors[key]

    def __setitem__(self, key, coordinates):
        if len(coordinates) == self.dimension:  # Check if coordinates respect self.dimension.
            if isinstance(coordinates, np.ndarray):  # Check if coordinates are numpy arrays.

                self.vectors[key] = coordinates

            else:
                raise TypeError('`coordinates` argument should be a numpy array.')
        else:
            raise ValueError('`coordinates` argument length should be equal to {0} not {1}'
                             .format(self.dimension, len(coordinates)))

    def keys(self):
        return list(self.vectors.keys())

    def values(self):
        return list(self.vectors.values())

    def items(self):
        return list(self.vectors.items())

    def most_similar(self, coordinates, best=5):
        """
        Return the `best` most similar dict entries of `coordinates`.
        """

        similarity_dict = dict()

        for key, key_coordinates in self.items():
            similarity_dict[key] = 1 - spatial.distance.cityblock(key_coordinates, coordinates)  # 1 - distance = similarity

        return Counter(similarity_dict).most_common(best)

    def apply_translation(self, from1, to1, from2, best=5):
        return self.most_similar(to1 - from1 + from2, best)

    def reduce(self, to_dimension=2):
        """
        Perform a SVD reduction on the dict.

        :param to_dimension: New dict dimension
        :return: A new dict with values that respect `to_dimension` arg
        """
        raw_coordinates = np.array([coordinates for coordinates in self.values()])

        SVD_model = TruncatedSVD(n_components=to_dimension)
        reduced_coordinates = SVD_model.fit_transform(raw_coordinates)

        reduced_dict = VectorDictionary(dimension=to_dimension)
        for index, key in enumerate(self.keys()):
            reduced_dict[key] = reduced_coordinates[index]

        return reduced_dict